/**
 * Instrumentation Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.log.records;

import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.StringTokenizer;

import net.yale.cs.p4p.p2p.azureus.log.XFerLogPlugin;
import net.yale.cs.p4p.p2p.logs.DownloadLogs;

import org.gudy.azureus2.plugins.PluginConfig;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadException;
import org.gudy.azureus2.plugins.download.DownloadScrapeResult;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.torrent.Torrent;

public class PluginDownloadLogs extends DownloadLogs {

	private static MessageDigest HASH_ALG = null;
	static {
		try {
			HASH_ALG = MessageDigest.getInstance("SHA-1");
		} catch (NoSuchAlgorithmException e) {
			System.err.println("Failed to locate hash algorithm: " + e);
		}
	}

	/* Get the hash of an input string */
	private static byte[] getHash(byte[] a) {
		if (HASH_ALG == null)
			return a;

		synchronized (HASH_ALG) {
			HASH_ALG.reset();
			return HASH_ALG.digest(a);
		}
	}

	/* Trims a hash array the desired length */
	private static byte[] trimHash(byte[] hash, int length) {
		if (hash.length == 0)
			throw new IllegalArgumentException("hash must not be empty");

		/* Ensure length is in range */
		length = Math.max(1, Math.min(length, hash.length));
		if (length == hash.length)
			return hash;

		/* Return subset of hash */
		byte[] newHash = new byte[length];
		System.arraycopy(hash, 0, newHash, 0, length);
		return newHash;
	}

	/* Zeroes-out all but leading n bits of IP address */
	private static byte[] trimAddress(byte[] addr, int len) {
		if (len >= addr.length * 8)
			return addr;

		int partialByteIdx = len / 8;

		/* Zero out partial bits */
		int zeroBits = 8 - (len % 8);
		addr[partialByteIdx] = (byte)(addr[partialByteIdx] & (0xFF << zeroBits));

		/* Zero out remaining bits */
		for (int i = partialByteIdx + 1; i < addr.length; ++i)
			addr[i] = 0;

		return addr;
	}

	private static String trimAddress(XFerLogPlugin plugin, String address) {
		try {
			InetAddress addr = InetAddress.getByName(address);
			if (addr instanceof Inet4Address)
				return InetAddress.getByAddress(trimAddress(addr.getAddress(), plugin.getConfig().getCfgAddrKeepBitsV4())).getHostAddress();
			else if (addr instanceof Inet6Address)
				return InetAddress.getByAddress(trimAddress(addr.getAddress(), plugin.getConfig().getCfgAddrKeepBitsV6())).getHostAddress();
			else
				return address;
		} catch (UnknownHostException e) {
			return address;
		}
	}

	private static String getPeerKind(Peer peer) {
		Object kind = peer.getUserData("LOG_KIND");
		return (kind != null && kind instanceof String) ? (String)kind : "Unknown";
	}

	private static Tags getPeerTags(Peer peer) {
		Object tags = peer.getUserData("LOG_TAGS");
		Tags result = new Tags();
		if (tags == null || !(tags instanceof String))
			return result;

		StringTokenizer t = new StringTokenizer((String)tags, ",");
		while (t.hasMoreTokens())
			result.add(t.nextToken());
		return result;
	}

	private static void setCommonParams(LogRecord rec, XFerLogPlugin plugin) {
		rec.setLogVersion(plugin.getPluginInterface().getPluginVersion());
		rec.setClientId(plugin.getPluginInterface().getUtilities().getSecurityManager().getIdentity());
	}

	private static void setTorrentParams(LogRecord rec, XFerLogPlugin plugin, Download download) {
		Torrent torrent = download.getTorrent();
		if (torrent != null) {
			rec.setTorrentId(trimHash(getHash(torrent.getHash()), plugin.getConfig().getCfgHashKeepBytes()));
			rec.setFileSize(torrent.getSize());
		}
		
		DownloadScrapeResult scrapeResult = download.getLastScrapeResult();
		if (scrapeResult != null) {
			rec.setNumLeechers(scrapeResult.getNonSeedCount());
			rec.setNumSeeders(scrapeResult.getSeedCount());
		}
	}

	private static boolean noSelfPortAvailable = false;
	private static int getSelfPort(XFerLogPlugin plugin) {
		/* Handle older clients that don't support this */
		if (noSelfPortAvailable)
			return 0;

		try {
			return plugin.getPluginInterface().getPluginconfig().getCoreIntParameter(PluginConfig.CORE_PARAM_INT_INCOMING_TCP_PORT);
		} catch (Throwable e) {
			noSelfPortAvailable = true;
			return 0;
		}
	}

	private static void setSelfParams(LogRecord rec, XFerLogPlugin plugin, Download download) {
		rec.setSelfPort(getSelfPort(plugin));
		if (download != null) {
			rec.setSelfId(download.getDownloadPeerId());
			if (download.getPeerManager() != null)
				rec.setSelfType(download.getPeerManager().isSeeding() ? NodeType.NODE_SEED : NodeType.NODE_LEECH);
		}
	}
	
	private static void setPeerParams(LogRecord rec, XFerLogPlugin plugin, Peer peer, PeerTransferStats stats) {
		rec.setPeerId(peer.getId());
		rec.setPeerIp(trimAddress(plugin, peer.getIp()));
		rec.setPeerPort(peer.getPort());
		rec.setPeerKind(getPeerKind(peer));
		rec.setPeerTags(getPeerTags(peer));
		rec.setPeerType(peer.isSeed() ? NodeType.NODE_SEED : NodeType.NODE_LEECH);
		if (stats != null) {
			rec.setPeerUploadCap(stats.getMaxPeerUploadRate());
			rec.setPeerDownloadCap(stats.getMaxPeerDownloadRate());
		}
	}

	public static LogRecord MakeLogRecordConfigChanged(XFerLogPlugin plugin) {
		LogRecord rec = new LogRecord();
		rec.setEvent(Event.EVT_CONFIG_CHANGED);
		setCommonParams(rec, plugin);
		setSelfParams(rec, plugin, null);
		rec.setComponents(plugin.getComponents());
		rec.setStartTime(System.currentTimeMillis());
		rec.setEndTime(rec.getStartTime());
		return rec;
	}

	public static LogRecord MakeLogRecordDownloadStart(XFerLogPlugin plugin, Download download) {
		LogRecord rec = new LogRecord();
		rec.setEvent(Event.EVT_DLOAD_STATS);
		setCommonParams(rec, plugin);
		setTorrentParams(rec, plugin, download);
		setSelfParams(rec, plugin, download);
		rec.setStartTime(System.currentTimeMillis());
		rec.setBytesReceived(download.getStats().getDownloaded());
		rec.setBytesSent(download.getStats().getUploaded());
		rec.setBytesDiscarded(download.getStats().getDiscarded());
		return rec;
	}

	public static LogRecord MakeLogRecordDownloadEnd(LogRecord start, Download download) {
		start.setEndTime(System.currentTimeMillis());
		start.setBytesReceived(download.getStats().getDownloaded() - start.getBytesReceived());
		start.setBytesSent(download.getStats().getUploaded() - start.getBytesSent());
		start.setBytesDiscarded(download.getStats().getDiscarded() - start.getBytesDiscarded());
		return start;
	}

	public static LogRecord MakeLogRecordTransferEvent(XFerLogPlugin plugin, Event event, Peer peer, PeerTransferStats stats) {
		Download download = null;
		try {
			download = peer.getManager().getDownload();
		} catch (DownloadException e) { /* Ignore */ }

		LogRecord rec = new LogRecord();
		rec.setEvent(event);
		setCommonParams(rec, plugin);
		setTorrentParams(rec, plugin, download);
		setSelfParams(rec, plugin, download);
		setPeerParams(rec, plugin, peer, stats);
		rec.setStartTime(stats.getStartTime());
		rec.setEndTime(stats.getEndTime());
		rec.setBytesReceived(stats.getBytesReceived());
		rec.setBytesSent(stats.getBytesSent());
		rec.setBytesDiscarded(stats.getBytesDiscarded());
		return rec;
	}

	public static LogRecord MakeLogRecordPeerEvent(XFerLogPlugin plugin, Event event, Download download, Peer peer) {
		LogRecord rec = new LogRecord();
		rec.setEvent(event);
		setCommonParams(rec, plugin);
		setTorrentParams(rec, plugin, download);
		setSelfParams(rec, plugin, download);
		setPeerParams(rec, plugin, peer, null);
		rec.setStartTime(System.currentTimeMillis());
		rec.setEndTime(rec.getStartTime());
		return rec;
	}

}
